#!/usr/bin/env python3

from typing import Any

from ..extension import Extension
from ..tree import SourceNode, SourceIndex

"""The extension for generating breadcrumb informatioon."""


class Breadcrumbs(Extension):
    def __init__(self, config: dict[str, Any]):
        """
        Initialize the Breadcrumbs extension.

        Args:
            config (dict[str, Any]): Configuration variables for this extension.
        """
        super(Breadcrumbs, self).__init__(config)
        self._root_nodes: list[SourceNode] = []
        self._raw_targets: list[tuple[SourceNode, str]] = []
        self._parents: dict[SourceNode, SourceNode] = {}

    def loaded_metadata(self, ctxt: Any, node: SourceNode) -> None:
        """
        Called after the metadata for a node has been loaded.  Should be overridden by derived classes.

        Args:
            ctxt (Context): Context for the operation.
            node (SourceNode): Node that's just had its metadata loaded.
        """
        if node.publish:
            target_str = node.metadata.get("parent", None)
            if target_str:
                target_str = str(target_str).strip()
                if target_str[:2] == '[[' and target_str[-2:] == ']]':
                    target_str = target_str[2:-2].strip()
                    tmp = target_str.split('|')
                    if len(tmp) > 1:
                        target_str = tmp[0].strip()
                if target_str == '.':
                    self._root_nodes.append(node)
                else:
                    self._raw_targets.append((node, target_str))

    def indexed(self, ctxt: Any, index: SourceIndex) -> None:
        """
        Called after the index for the nodes has been generated.

        Args:
            ctxt (Context): Context for the operation.
            index (SourceIndex): The index to all nodes.
        """
        for node, target in self._raw_targets:
            target_node, _ = index.lookup(node, target)
            if target_node:
                self._parents[node] = target_node

    def _crumb_name(self, node:SourceNode) -> str:
        """
        Returns the "crumb name" for a node.

        Args:
            node (SourceNode): The source node to return the "crumb name" of.
        """
        if node in self._root_nodes:
            return '.'
        return node.metadata.get("crumb_name", node.page_title)

    def supply_template_vars(self, ctxt: Any, node: SourceNode) -> dict[str, Any] | None:
        """
        Called to supply additional template variables to the template to be rendered.

        Args:
            ctxt (Context): Context for the operation.
            node (SourceNode): The node currently being processed for template rendering.

        Returns:
            dict[str, Any]: If not ``None``, contains additional template variables to be added before
                the template is rendered.
        """
        breadcrumbs: list[dict[str, str]] = [{"name": self._crumb_name(node), "link": ""}]
        cur_node = node
        while cur_node not in self._root_nodes:
            cur_node = self._parents.get(cur_node, None)
            if not cur_node:
                return None
            breadcrumbs.insert(0, {"name": self._crumb_name(cur_node),
                                   "link": self.link_target(ctxt, node, cur_node)})
        return {"breadcrumbs": breadcrumbs}
