#!/usr/bin/env python

"""The files and directories of the source Obsidian vault."""

from pathlib import Path
from typing import Any
from urllib.parse import quote as urlquote

import logging
import markdown
import yaml

logger = logging.getLogger(__name__)

"""The paths that are always to be ignored."""
STATIC_IGNORE = [
    '.obsidian',
    '.dragonglass'
]

"""The path matching Markdown files."""
MARKDOWN_PAT = '*.md'


class SourceNode:
    """
    Represents a file or directory in the source Obsidian vault.

    Attributes:
        metadata (dict[str, Any]): The metadata from the current node.
        text (str): The parsed HTML text of the current node.
        backlinks (set[SourceNode]): Set of nodes that link to this one.
    """
    def __init__(self, root: Path, path: Path, is_dir: bool) -> None:
        """
        Initialize the SourceNode.

        Args:
            root (Path): Root path the node lives under.
            path (Path): Relative path from the root to this node.
            is_dir (bool): ``True`` if the node is a directory, ``False`` if not.
        """
        self._root = root
        self._path = path
        self._is_dir = is_dir
        self._is_md = path.match(MARKDOWN_PAT)
        self._tags: dict[str, str] = {}
        self.metadata: dict[str, Any] = {}
        self.text: str | None = None
        self.backlinks: set[Any] = set()

    def __str__(self) -> str:
        """Returns the string representation of the nmode."""
        return f"SourceNode({self._path}, {self._is_dir}) [is_md={self._is_md}]"

    @property
    def is_dir(self) -> bool:
        """Returns ``True`` if this node refers to a directory, ``False`` if it refers to an ordinary file."""
        return self._is_dir

    @property
    def is_md(self) -> bool:
        """Returns ``True`` if this node refers to a Markdown file, ``False`` if not."""
        return self._is_md

    @property
    def path(self) -> Path:
        """Returns the path, relative to the root path, of this node."""
        return self._path

    @property
    def root_path(self) -> Path:
        """Returns the root path this node is under."""
        return self._root

    @property
    def page_title(self) -> str:
        """Returns the standard page title for this node."""
        return self.metadata.get("title", self._path.stem)

    @property
    def publish(self) -> bool:
        """Returns the "publish" flag for a page."""
        return self.metadata.get("publish", True) if self._is_md else True

    def target_file(self, dest_dir: Path) -> Path:
        """
        Computes the path of the target file as it will be written to the destination directory.

        Args:
            dest_dir (Path): The destination directory.

        Returns:
            Path: The path of the target file to be written.
        """
        p = self._path
        if self._is_md:
            p = p.with_suffix('.html')
        return dest_dir / p

    def link_target(self, prefix: str = "/", rel_to: Any = None) -> str:
        """
        Returns the link target string for this node.

        Args:
            prefix (str): The prefix to apply to all absolute URLs.  Default is "/".
            rel_to (SourceNode): If this is not ``None``, the path is to be computed as a relative path to the
                specified node.

        Returns:
            str: The link target, as a string.
        """
        xpath = self._path.with_suffix('.html') if self._is_md else self._path
        if rel_to is not None:
            xpath = self._root / xpath
            rel_path = rel_to.root_path / rel_to.path
            logger.debug(f"*** Computing path of {xpath} relative to {rel_path}")
            return urlquote(xpath.relative_to(rel_path.parent, walk_up=True).as_posix())
        return urlquote(prefix + xpath.as_posix())

    def stash_tag(self, tagname: str) -> None:
        """
        Add a tag to the set of tags in this node.

        Args:
            tagname (str): The tag name to be added.
        """
        rtag = tagname.lower()
        if rtag not in self._tags:
            self._tags[rtag] = tagname

    def load_metadata(self) -> None:
        """
        Loads the metadata for this particular node and saves it in the "metadata" attribute.

        The metadata is YAML at the beginning of the Markdown file, and is set via Obsidian "properties."
        """
        if self._is_md and not self._is_dir:
            with open(self._root / self._path, "r", encoding="utf-8") as f:
                cur_line = f.readline()
                if cur_line == '---\n':
                    metalines: list[str] = []
                    cur_line = f.readline()
                    while cur_line != '---\n':
                        metalines.append(cur_line)
                        cur_line = f.readline()
                    self.metadata = yaml.full_load(''.join(metalines))
            # Load up the initial tags.
            for tag in self.metadata.get("tags", []):
                self.stash_tag(tag)

    def parse_markdown(self, markdown_parser: markdown.Markdown) -> None:
        """
        Parses the Markdown in this file into HTML, saving the HTML into the "text" attribute.

        Args:
            markdown_parser (markdown.Markdown): Instance of the Markdown parser to use in the parsing.
        """
        if self._is_md and not self._is_dir:
            markdown_parser.reset()
            with open(self._root / self._path, "r", encoding="utf-8") as f:
                self.text = markdown_parser.convert(f.read())

    def make_vars(self) -> dict[str, Any]:
        """
        Creates the template variables to use for this node.

        Returns:
            dict[str, Any]: The template variables, as a dict.
        """
        return {
            "text": self.text,
            "title": self.page_title,
            "description": self.metadata.get("description", ""),
            "tags": [self._tags[n] for n in sorted(self._tags.keys())]
        }


def generate_list(ctxt: Any, source_root: Path) -> list[SourceNode]:
    """
    Generates the list of source nodes from the source path (Obsidian vault).

    Args:
        source_root (Path): The root directory (Obsidian vault) to load nodes from.

    Returns:
        list[SourceNode]: The list of nodes (files and directories) under the root, as a preorder traversal.
    """
    nodes: list[SourceNode] = []
    dirs = [source_root]
    while len(dirs) > 0:
        current_dir = dirs.pop(0)
        for child in current_dir.iterdir():
            rchild = child.relative_to(source_root)
            add_me = True
            for pat in STATIC_IGNORE:
                if rchild.match(pat):
                    add_me = False
                    break
            if add_me and ctxt.file_ignored(rchild):
                add_me = False
            if add_me:
                nodes.append(SourceNode(source_root, rchild, child.is_dir()))
                if child.is_dir():
                    dirs.append(child)
    return nodes


class SourceIndex:
    """
    The index of all nodes in the Obsidian vault.  Nodes are listed by their name, by their relative path name,
    and by any aliases specified in their metadata.
    """
    def __init__(self, nodelist: list[SourceNode]) -> None:
        """
        Initialize the source index.

        Args:
            nodelist (list[SourceNode]): The list of nodes to be indexed.
        """
        self._bypath: dict[str, SourceNode] = {}
        self._byname: dict[str, SourceNode] = {}
        self._byalias: dict[str, SourceNode] = {}
        for node in nodelist:
            if node.is_dir:
                continue
            self._bypath[node.path.as_posix()] = node
            if node.is_md:
                tmp = node.path.with_suffix('')
                key = tmp.name
                if key not in self._byname:
                    self._byname[key] = node
                self._byname[tmp.as_posix()] = node
                if node.metadata:
                    aliases: list[str] = node.metadata.get('aliases', [])
                    for alias in aliases:
                        if alias not in self._byalias:
                            self._byalias[alias] = node
            else:
                key = node.path.name
                if key not in self._byname:
                    self._byname[key] = node
                self._byname[node.path.as_posix()] = node

    def lookup(self, from_node: SourceNode, reference: str) -> tuple[SourceNode | None, str | None]:
        """
        Looks up a reference in the index and returns the corresponding node if it exists.

        Args:
            from_node (SourceNode): The "current" node; used to look up a node by relative path.
            reference (str): The reference to look up.

        Returns:
            SourceNode: The node that was found, or ``None`` if the node was not found.
            str: Indicates whether the match was on "NAME", "ALIAS", or "PATH".  Returns ``None``
                if the node was not found.
        """
        if reference in self._byname:
            return self._byname[reference], 'NAME'
        elif reference in self._byalias:
            return self._byalias[reference], 'ALIAS'
        else:
            new_path = (from_node.path.parent / reference).as_posix()
            if new_path in self._bypath:
                return self._bypath[new_path], 'PATH'
            return None, None
